@@ -4,24 +4,46 @@ use Moose;
 with 'MooseX::Emulate::Class::Accessor::Fast';
 
 use Data::Dump;
-use Class::MOP ();
+use Moose::Util 'find_meta';
+use Carp qw/ cluck /;
 
-our %LEVELS = ();
+our %LEVELS = (); # Levels stored as bit field, ergo debug = 1, warn = 2 etc
+our %LEVEL_MATCH = (); # Stored as additive, thus debug = 31, warn = 30 etc
 
 has level => (is => 'rw');
 has _body => (is => 'rw');
 has abort => (is => 'rw');
+has _psgi_logger => (is => 'rw', predicate => '_has_psgi_logger', clearer => '_clear_psgi_logger');
+has _psgi_errors => (is => 'rw', predicate => '_has_psgi_errors', clearer => '_clear_psgi_errors');
+
+sub clear_psgi {
+    my $self = shift;
+    $self->_clear_psgi_logger;
+    $self->_clear_psgi_errors;
+}
+
+sub psgienv {
+    my ($self, $env) = @_;
+
+    $self->_psgi_logger($env->{'psgix.logger'}) if $env->{'psgix.logger'};
+    $self->_psgi_errors($env->{'psgi.errors'}) if $env->{'psgi.errors'};
+}
+
 
 {
     my @levels = qw[ debug info warn error fatal ];
 
-    my $meta = Class::MOP::get_metaclass_by_name(__PACKAGE__);
-    for ( my $i = 0 ; $i < @levels ; $i++ ) {
+    my $meta = find_meta(__PACKAGE__);
+    my $summed_level = 0;
+    for ( my $i = $#levels ; $i >= 0 ; $i-- ) {
 
         my $name  = $levels[$i];
+
         my $level = 1 << $i;
+        $summed_level |= $level;
 
         $LEVELS{$name} = $level;
+        $LEVEL_MATCH{$name} = $summed_level;
 
        $meta->add_method($name, sub {
             my $self = shift;
@@ -42,7 +64,9 @@ around new => sub {
     my $orig = shift;
     my $class = shift;
     my $self = $class->$orig;
+
     $self->levels( scalar(@_) ? @_ : keys %LEVELS );
+
     return $self;
 };
 
@@ -55,7 +79,7 @@ sub levels {
 sub enable {
     my ( $self, @levels ) = @_;
     my $level = $self->level;
-    for(map { $LEVELS{$_} } @levels){
+    for(map { $LEVEL_MATCH{$_} } @levels){
       $level |= $_;
     }
     $self->level($level);
@@ -70,8 +94,12 @@ sub disable {
     $self->level($level);
 }
 
+our $HAS_DUMPED;
 sub _dump {
     my $self = shift;
+    unless ($HAS_DUMPED++) {
+        cluck("Catalyst::Log::_dump is deprecated and will be removed. Please change to using your own Dumper.\n");
+    }
     $self->info( Data::Dump::dump(@_) );
 }
 
@@ -79,10 +107,17 @@ sub _log {
     my $self    = shift;
     my $level   = shift;
     my $message = join( "\n", @_ );
-    $message .= "\n" unless $message =~ /\n$/;
-    my $body = $self->_body;
-    $body .= sprintf( "[%s] %s", $level, $message );
-    $self->_body($body);
+    if ($self->can('_has_psgi_logger') and $self->_has_psgi_logger) {
+        $self->_psgi_logger->({
+                level => $level,
+                message => $message,
+            });
+    } else {
+        $message .= "\n" unless $message =~ /\n$/;
+        my $body = $self->_body;
+        $body .= sprintf( "[%s] %s", $level, $message );
+        $self->_body($body);
+    }
 }
 
 sub _flush {
@@ -98,16 +133,37 @@ sub _flush {
 
 sub _send_to_log {
     my $self = shift;
-    print STDERR @_;
+    if ($self->can('_has_psgi_errors') and $self->_has_psgi_errors) {
+        $self->_psgi_errors->print(@_);
+    } else {
+        print STDERR @_;
+    }
 }
 
+# 5.7 compat code.
+# Alias _body to body, add a before modifier to warn..
+my $meta = __PACKAGE__->meta; # Calling meta method here fine as we happen at compile time.
+$meta->add_method('body', $meta->get_method('_body'));
+my %package_hash; # Only warn once per method, per package.
+                  # I haven't provided a way to disable them, patches welcome.
+$meta->add_before_method_modifier('body', sub {
+    my $class = blessed(shift);
+    $package_hash{$class}++ || do {
+        warn("Class $class is calling the deprecated method Catalyst::Log->body method,\n"
+            . "this will be removed in Catalyst 5.81");
+    };
+});
+# End 5.70 backwards compatibility hacks.
+
 no Moose;
-__PACKAGE__->meta->make_immutable();
+__PACKAGE__->meta->make_immutable(inline_constructor => 0);
 
 1;
 
 __END__
 
+=for stopwords psgienv
+
 =head1 NAME
 
 Catalyst::Log - Catalyst Log Class
@@ -220,8 +276,8 @@ Is the log level active?
 
 =head2 abort
 
-Should Catalyst emit logs for this request? Will be reset at the end of 
-each request. 
+Should Catalyst emit logs for this request? Will be reset at the end of
+each request.
 
 *NOTE* This method is not compatible with other log apis, so if you plan
 to use Log4Perl or another logger, you should call it like this:
@@ -236,6 +292,21 @@ This protected method is what actually sends the log information to STDERR.
 You may subclass this module and override this method to get finer control
 over the log output.
 
+=head2 psgienv $env
+
+    $log->psgienv($env);
+
+NOTE: This is not meant for public consumption.
+
+Set the PSGI environment for this request. This ensures logs will be sent to
+the right place. If the environment has a C<psgix.logger>, it will be used. If
+not, we will send logs to C<psgi.errors> if that exists. As a last fallback, we
+will send to STDERR as before.
+
+=head2 clear_psgi
+
+Clears the PSGI environment attributes set by L</psgienv>.
+
 =head2 meta
 
 =head1 SEE ALSO
@@ -248,7 +319,7 @@ Catalyst Contributors, see Catalyst.pm
 
 =head1 COPYRIGHT
 
-This program is free software, you can redistribute it and/or modify
+This library is free software. You can redistribute it and/or modify
 it under the same terms as Perl itself.
 
 =cut
